#version 120

#define WATER_REFLECTIONS

//reflection - raytrace settings
const float stp = 1.0f;    // step multiplier (0.5f - 1.0f)
const float ref = 0.1f;    // refinement multiplier (< 1.0f)
const float inc = 2.0f;     // increment factor at each step (> 1.0f)
const int maxf = 6;         // number of refinements

varying vec4 texcoord;

varying vec4 lightVector;
varying vec3 sunlight;

varying vec3 caustics;

varying float transition_fading;
varying float fog_level;

uniform sampler2D gcolor;
uniform sampler2D gdepth;   // heavily bump water normal from gbuffer_water.fsh
uniform sampler2D gnormal;
uniform sampler2D composite;
uniform sampler2D gaux1;
uniform sampler2D gaux2;    // rain from gbuffer_weather.fsh
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D depthtex2;

uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;

uniform vec3 sunPosition;

uniform float far;
uniform float near;
uniform float viewWidth;
uniform float viewHeight;
uniform float rainStrength;

uniform int isEyeInWater;
uniform int worldTime;
uniform int fogMode;

float pw = 1.0f / viewWidth;
float ph = 1.0f / viewHeight;

float gp00 = gbufferProjection[0][0] * -0.5f;
float gp11 = gbufferProjection[1][1] * -0.5f;
float gp22 = gbufferProjection[2][2] * 0.5f;
float gp32 = gbufferProjection[3][2] * 0.5f;
float gp2232 = -gp22 - gp32;
float igp00 = -1.0f / gp00;
float igp11 = -1.0f / gp11;

vec3 proj(in vec3 pos) {
    return vec3(pos.x * gp00, pos.y * gp11, gp2232) / pos.z + 0.5f;
}

vec3 invproj(in vec3 p) {
    vec3 pos = p - 0.5f;
    float z = gp32 / (pos.z + gp22);
    return vec3(pos.x * igp00, pos.y * igp11, -1.0f) * z;
}

float cdist(in vec2 coord) {
    return distance(coord, vec2(0.5f)) * 2.0f;
}

vec4 raytrace(in vec3 eye, in vec3 refvector) {
    vec4 color = vec4(0.0f);
    float delta = stp;
    float sum = delta;
    int sr = 0;
    for(int i=0;i<30;i++) {
        vec3 offset = eye + sum * refvector;
        vec3 pos = proj(offset);
        if (pos.x < 0.0f || pos.x > 1.0f || pos.y < 0.0f || pos.y > 1.0f || pos.z < 0.0f || pos.z > 1.0f) break;
        vec3 match = invproj(vec3(pos.st, texture2D(depthtex1, pos.st).r));
        if (distance(offset, match) < delta * pow(sum, 0.1f) * 1.75f) {
            if (sr >= maxf) {
                float border = clamp(1.0f - pow(cdist(pos.st), 1.5f), 0.0f, 1.0f);
                color = texture2D(composite, pos.st);
                color.a = 0.9f;
                color.a *= border;
                break;
            }
            sr++;
            sum -= delta;
            delta *= ref;
        }
        delta *= inc;
        sum += delta;
    }
    return color;
}

float ld(in float depth) {
    return (2.0f * near) / (far + near - depth * (far - near));
}

//////////////////////////////main//////////////////////////////

void main() {
    float matflag = texture2D(gaux1, texcoord.xy).g;
    const float fland = 0.05f;
    const float fwater = 0.95f;
    const float freflect = 0.85f;
    float hand = 0.0f;
    
    if(matflag > 0.35f && matflag < 0.45f) {
        hand = 0.7f;    // scale rain effect on hand (0.1f - 1.0f)
    }
    
    vec3 fogclr = mix(gl_Fog.color.rgb, vec3(0.3f, 0.4f, 0.5f), 0.2f);
    vec3 eye = invproj(vec3(texcoord.st, texture2D(depthtex0, texcoord.st).r));
    vec4 color = texture2D(gcolor, texcoord.st);
    vec3 normal = texture2D(gnormal, texcoord.st).rgb * 2.0f - 1.0f;

    float ssao = 1.0f;
    if (matflag > fland && matflag < fwater && hand < 0.1f) {
		ssao = 0.0f;
		float range = clamp(distance(proj(eye + vec3(1.0f)).xy, texcoord.xy), 7.5f * pw, 60.0f * pw);
        float dist = ld(texture2D(depthtex0, texcoord.st).r);
		for (float i = 0.0f; i < 5.1f; i += 1.3f) {
			for (float j = 0.25f; j < 1.1f; j += 0.25f) {
				vec2 offset = vec2(cos(float(i)), sin(float(i))) * (float(j) / 4.0f) * range + texcoord.xy;
				vec3 offpos = invproj(vec3(offset, texture2D(depthtex0, offset).r));
				float angle = pow(min(1.0 - dot(normal, normalize(offpos - eye)), 1.0f), 2.0f);
				float ldist = pow(min(abs(ld(texture2D(depthtex0, offset).r) - dist) * 65.0f, 1.0f), 2.0f);
				ssao += pow(min(ldist + angle, 1.0f), 3.0f);
			}
		}
		ssao /= 16.0f;
	}
    color.rgb *= ssao * 0.5f + 0.5f;

    color.rgb += texture2D(gaux2, texcoord.xy).rgb * texture2D(gaux2, texcoord.xy).a * 0.5f * (1.0f - hand);	// draw rain

    if (isEyeInWater < 1) {
        vec3 neye = normalize(eye);
        vec3 nnormal = normalize(normal);
        if (matflag > freflect) {
#ifdef WATER_REFLECTIONS
            float fresnel = clamp(pow(1.0f + dot(nnormal, neye), 5.0f), 0.0f, 1.0f);
            vec3 fnormal = mix(normal, texture2D(gdepth, texcoord.st).rgb * 2.0f - 1.0f, fresnel);
            vec4 reflection = raytrace(eye, normalize(reflect(neye, normalize(fnormal))));
            reflection.rgb = mix(fogclr, reflection.rgb, reflection.a); // add sky/fog color to reflection
            float reflectionStrength = matflag * 5.0f - 4.0f;
            color.rgb += reflection.rgb * fresnel * reflectionStrength * (1.0f - rainStrength * 0.5f);
            //color.r = mod(-eye.z, 5.0f) > 0.95 * 5.0f? 1.0f : color.r;   // show lines on water
            color.rgb += color.a * reflection.rgb * reflectionStrength * 0.75f * fresnel;  // spec from composite.fsh here to remove shadows
            color.r *= 0.8f;    // set 80 percent red on water
#endif
        }
        if (matflag > fwater) {
            float f0 = 0.3f * max(dot(caustics, nnormal), 0.00001f); // basic caustics
            f0 *= step(f0, 0.02f);  // remove artifacts from flowing water
            color.g -= f0;
        }
        if (fogMode == 0 && hand < 0.1f) {
            vec4 worldpos = gbufferModelViewInverse * vec4(eye, 1.0f);
            float density = 96.0f / max(120.0f + worldpos.y, 52.7f) - 0.8f;
            float dist = (1.0f + rainStrength * 0.4f) * 180.0f / max(280.0f - length(eye) * fog_level * 215.0f / far, 122.0f) - 0.5f + rainStrength * 0.1f;
            float fog = clamp(mix(density, dist, 0.5f), 0.0f, 1.0f);
            fog *= 1.0f - 0.4f * step(matflag, fland) * (1.0f - rainStrength * 0.9f);  // reduce fog in sky
            float volumetric_cone = pow(max(dot(neye, lightVector.xyz), 0.0f), 5.0f) * lightVector.a;
            fogclr += sunlight * transition_fading * volumetric_cone * (1.0f - rainStrength * 0.9f);   // inject moon color into the fog
            color.rgb = mix(color.rgb, fogclr, fog);
            //float range = length(eye);
            //color.r = (range > 175.0f && range < 176.0f) ? 1.0f : color.r;
            //color.g = (worldpos.y > -52.0f && worldpos.y < -51.0f) ? 1.0f : color.g;
            //color.b = (worldpos.y > -0.1f && worldpos.y < 0.0f) ? 1.0f : color.b;
        }
    }
    else {  // eye in water
        float depth_diff = clamp(pow(ld(texture2D(depthtex2, texcoord.st).r) * 3.4f, 2.0f), 0.0f, 1.0f);
        color.rgb = mix(color.rgb * sunlight * vec3(0.2f, 0.3f, 0.4f), vec3(0.1f, 0.15f, 0.2f), depth_diff);
    }

    //calculate sun occlusion at vec2(0.0f)
    float visiblesun = 0.0f;
    if (texcoord.x < pw && texcoord.y < ph) {
        vec4 tpos = vec4(sunPosition, 1.0f) * gbufferProjection; // a transposed projection
        vec2 lightPos = (tpos.xy / tpos.z) * 0.5f + 0.5f;
        if (lightPos.x < 1.1f && lightPos.x > -0.1f && lightPos.y < 1.1f && lightPos.y > -0.1f) {
            for (int i = 0; i < 9;i++) {
                for (int j = 0; j < 9 ;j++) {
                    if (texture2D(gaux1, lightPos + vec2(pw * (float(i) - 4.0f) * 12.0f, ph * (float(j) - 4.0f) * 12.0f)).g < fland) visiblesun += 1.0f;
                }
            }
            visiblesun /= 81.0f;
            visiblesun *= 1.0f - rainStrength * 0.9f;
            visiblesun *= clamp((13050.0f - float(worldTime)) * 0.002f, 0.0f, 1.0f) + clamp((float(worldTime) - 22950.0f) * 0.002f, 0.0f, 1.0f);
        }
    }
    
/* DRAWBUFFERS:0 */
    gl_FragData[0] = vec4(color.rgb, visiblesun);    // gcolor in final.fsh
}

//////////////////////////////main//////////////////////////////
